# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import jsonschema
import logging
import math

from aws_sso_lib import lookup
from aws_sso_lib.format import format_account_id

from . import utils, cfn_yaml_tags

LOGGER = logging.getLogger(__name__)

class ConfigError(Exception):
    pass

def _get_value(dct, keys, ensure_list=False, type=None):
    if not isinstance(keys, list):
        keys = [keys]
    found_key = None
    found_value = [] if ensure_list else None
    for key in keys:
        if key in dct:
            value = dct[key]
            if value is None:
                continue
            if ensure_list and not isinstance(value, list):
                value = [value]
            found_key = key
            found_value = value
            break
    if type is not None:
        if ensure_list:
            found_value = [type(v) for v in found_value]
        elif found_value is not None:
            found_value = type(found_value)
    return found_key, found_value

class GenerationConfig:
    DEFAULT_MAX_RESOURCES_PER_TEMPLATE = 500
    DEFAULT_MAX_CONCURRENT_ASSIGNMENTS = 20
    DEFAULT_NUM_CHILD_STACKS = None

    def __init__(self,
            ids: lookup.Ids,
            principal_name_fetcher=None,
            permission_set_name_fetcher=None,
            target_name_fetcher=None):
        self.ids = ids
        self._max_resources_per_template = None
        self._max_concurrent_assignments = None
        self._max_assignments_allocation = None
        self._num_child_stacks = None

        self._default_session_duration = None

        self.principal_name_fetcher = principal_name_fetcher
        self.permission_set_name_fetcher = permission_set_name_fetcher
        self.target_name_fetcher = target_name_fetcher

    def __str__(self):
        return str({
            "max_resources_per_template": self.max_resources_per_template,
            "max_concurrent_assignments": self.max_concurrent_assignments,
            "num_child_stacks": self.num_child_stacks,
            "default_session_duration": self.default_session_duration,
            "internal": {
                "max_resources_per_template": self._max_resources_per_template,
                "max_concurrent_assignments": self._max_concurrent_assignments,
                "max_assignments_allocation": self._max_assignments_allocation,
                "num_child_stacks": self._num_child_stacks,
                "default_session_duration": self._default_session_duration,
            },
        })

    def copy(self):
        obj = self.__class__(
            self.ids,
            principal_name_fetcher=self.principal_name_fetcher,
            permission_set_name_fetcher=self.permission_set_name_fetcher,
            target_name_fetcher=self.target_name_fetcher
        )
        obj.set(
            max_resources_per_template=self._max_resources_per_template,
            max_concurrent_assignments=self._max_concurrent_assignments,
            max_assignments_allocation=self._max_assignments_allocation,
            num_child_stacks=self._num_child_stacks,
            default_session_duration=self._default_session_duration,
        )
        return obj

    @property
    def max_resources_per_template(self):
        if self._max_resources_per_template is None or self._max_resources_per_template < 1:
            return self.DEFAULT_MAX_RESOURCES_PER_TEMPLATE
        else:
            return self._max_resources_per_template

    @property
    def max_concurrent_assignments(self):
        if self._max_concurrent_assignments is None or self._max_concurrent_assignments < 1:
            return self.DEFAULT_MAX_CONCURRENT_ASSIGNMENTS
        else:
            return self._max_concurrent_assignments

    @property
    def num_child_stacks(self):
        num_stacks_1 = None
        if self._max_assignments_allocation is not None and self._max_assignments_allocation >= 1:
            num_stacks_1 = math.ceil(self._max_assignments_allocation / self.max_resources_per_template)

        num_stacks_2 = None
        if self._num_child_stacks is not None and self._num_child_stacks >= 0:
            num_stacks_2 = self._num_child_stacks

        if num_stacks_1 is not None and num_stacks_2 is not None:
            return max(num_stacks_1, num_stacks_2)
        elif num_stacks_1 is not None:
            return num_stacks_1
        elif num_stacks_2 is not None:
            return num_stacks_2
        else:
            return self.DEFAULT_NUM_CHILD_STACKS

    def get_max_number_of_child_stacks(self, num_resources):
        num_child_stacks = self.num_child_stacks
        if num_child_stacks is not None:
            return num_child_stacks
        return math.ceil(num_resources / self.max_resources_per_template)

    @property
    def default_session_duration(self):
        if not self._default_session_duration:
            return None
        else:
            return self._default_session_duration

    def set(self,
            max_resources_per_template=None,
            max_concurrent_assignments=None,
            max_assignments_allocation=None,
            num_child_stacks=None,
            default_session_duration=None,
            overwrite=False):

        if self._max_resources_per_template is None or (max_resources_per_template is not None and overwrite):
            self._max_resources_per_template = max_resources_per_template

        if self._max_concurrent_assignments is None or (max_concurrent_assignments is not None and overwrite):
            self._max_concurrent_assignments = max_concurrent_assignments

        if self._max_assignments_allocation is None or (max_assignments_allocation is not None and overwrite):
            self._max_assignments_allocation = max_assignments_allocation

        if self._num_child_stacks is None or (num_child_stacks is not None and overwrite):
            self._num_child_stacks = num_child_stacks

        if default_session_duration and (self._default_session_duration is None or overwrite):
            self._default_session_duration = default_session_duration

    def load(self, data, overwrite=False):
        max_resources_per_template = _get_value(data, ["MaxResourcesPerTemplate"], type=int)[1]
        max_concurrent_assignments = _get_value(data, ["MaxConcurrentAssignments"], type=int)[1]
        max_assignments_allocation = _get_value(data, ["MaxAssignmentsAllocation"], type=int)[1]
        num_child_stacks = _get_value(data, ["NumChildStacks", "NumChildTemplates"], type=int)[1]
        default_session_duration = _get_value(data, ["DefaultSessionDuration"], type=int)[1]

        self.set(
            max_resources_per_template=max_resources_per_template,
            max_concurrent_assignments=max_concurrent_assignments,
            max_assignments_allocation=max_assignments_allocation,
            num_child_stacks=num_child_stacks,
            default_session_duration=default_session_duration,
            overwrite=overwrite
        )

class Config:
    def __init__(self, data=None, resource_properties=None, resource_name_prefix=None):
        self.instance = None

        self.groups = []
        self.users = []

        self.permission_sets = []

        self.ous = []
        self.recursive_ous = []
        self.accounts = []

        self.assignment_group_name = None

        self.resource_name_prefix = resource_name_prefix

        if data:
            self.load(data)
        if resource_properties:
            self.load_resource_properties(resource_properties)

    def load(self, data):
        self.assignment_group_name = _get_value(data, ["AssignmentGroupName"])[1]

        self.instance = _get_value(data, ["Instance", "InstanceArn", "InstanceARN"])[1]

        def get(names):
            return _get_value(data, names, ensure_list=True)[1]

        self.groups.extend(get(["Groups", "Group"]))
        self.users.extend(get(["Users", "User"]))

        self.permission_sets.extend(get(["PermissionSet", "PermissionSetArn", "PermissionSets", "PermissionSetArns"]))

        self.ous.extend(get(["OUs", "Ous", "OU", "Ou"]))
        self.recursive_ous.extend(get(["RecursiveOUs", "RecursiveOus", "RecursiveOU", "RecursiveOu"]))
        for account in get(["Accounts", "Account"]):
            self.accounts.append(format_account_id(account))

    def load_resource_properties(self, resource_properties):
        data = {}

        name = _get_value(resource_properties, ["Name"])[1]
        if name is not None and isinstance(name, str):
            data["AssignmentGroupName"] = resource_properties["Name"]

        instance = _get_value(resource_properties, ["Instance", "InstanceArn", "InstanceARN"])[1]
        if instance is not None:
            data["Instance"] = instance

        principals = _get_value(resource_properties, ["Principal", "Principals"], ensure_list=True)[1]
        for principal_entry in principals:
            principal_type = _get_value(principal_entry, ["Type", "PrincipalType"])[1]
            principal_ids = _get_value(principal_entry, ["Id", "PrincipalId", "Ids", "PrincipalIds"], ensure_list=True)[1]
            if principal_type.upper() == "GROUP":
                config_key = "Groups"
            elif principal_type.upper() == "USER":
                config_key = "Users"
            else:
                raise ValueError(f"Invalid principal type: {principal_type}")
            if config_key not in data:
                data[config_key] = []
            data[config_key].extend(principal_ids)

        permission_sets = _get_value(resource_properties, ["PermissionSet", "PermissionSetArn", "PermissionSets", "PermissionSetArns"], ensure_list=True)[1]
        data["PermissionSets"] = permission_sets

        targets = _get_value(resource_properties, ["Target", "Targets"], ensure_list=True)[1]
        for target_entry in targets:
            target_type = _get_value(target_entry, ["Type", "TargetType"])[1]
            target_ids = _get_value(target_entry, ["Id", "TargetId", "Ids", "TargetIds"], ensure_list=True)[1]
            if target_type.upper() == "AWS_OU":
                if target_entry.get("Recursive", False):
                    config_key = "RecursiveOus"
                else:
                    config_key = "Ous"
            elif target_type.upper() == "AWS_ACCOUNT":
                config_key = "Accounts"
            else:
                raise ValueError(f"Invalid target type: {target_type}")
            if config_key not in data:
                data[config_key] = []
            data[config_key].extend(target_ids)

        import yaml

        self.load(data)

FUNC_SCHEMA = {
    "type": "object",
    "patternProperties": {
        "Ref|Fn::.*": {},
    },
    "minProperties": 1,
    "maxProperties": 1,
}

def _opt_func(schema):
    return {
        "oneOf": [
            schema,
            FUNC_SCHEMA,
        ]
    }

def _opt_list(schema, func=False):
    if func:
        return {
            "oneOf": [
                schema,
                {
                    "type": "array",
                    "items": {
                        "oneOf": [
                            schema,
                            FUNC_SCHEMA,
                        ]
                    }
                },
                FUNC_SCHEMA,
            ]
        }
    else:
        return {
            "oneOf": [
                schema,
                {
                    "type": "array",
                    "items": schema,
                },
            ]
        }

PRINCIPAL_SCHEMA = {
    "type": "object",
    "patternProperties": {
        "(Principal)?Type": _opt_func({
            "type": "string",
            "enum": ["GROUP", "USER"],
        }),
        "(Principal)?Id(s)?": _opt_list({"type": "string"}, func=True),
    }
}

PERMISSION_SET_SCHEMA = _opt_func({
    "type": "string",
})

TARGET_SCHEMA = {
    "type": "object",
    "patternProperties": {
        "(Target)?Type": _opt_func({
            "type": "string",
            "enum": ["AWS_OU", "AWS_ACCOUNT"],
        }),
        "(Target)?Id(s)?": _opt_list({"type": ["string", "integer"]}, func=True),
        "Recursive": {"type": "boolean"},
    }
}

RESOURCE_PROPERTY_SCHEMA = {
    "type": "object",
    "patternProperties": {
        "Name": _opt_func({"type": "string"}),
        "Instance(Arn|ARN)?": _opt_func({"type": "string"}),
        "Principal(s)?": _opt_list(PRINCIPAL_SCHEMA),
        "PermissionSet(Arn)?(s)?": _opt_list(PERMISSION_SET_SCHEMA),
        "Target(s)?": _opt_list(TARGET_SCHEMA),
        "UpdateNonce": _opt_func({"type": "string"}),
    },
    "additionalProperties": False,
}

def _check(properties, keys, required=True, parent="Resource"):
    found_keys = [k for k in keys if k in properties]
    if len(found_keys) > 1:
        raise ConfigError(f"{parent} must have only one of {', '.join(found_keys)}")
    if required and len(found_keys) == 0:
        raise ConfigError(f"{parent} must have one of {', '.join(keys)}")

def validate_resource(resource):
    resource = cfn_yaml_tags.to_json(resource)
    properties = resource.get("Properties", {})
    try:
        jsonschema.validate(
            schema=RESOURCE_PROPERTY_SCHEMA,
            instance=properties)
    except jsonschema.ValidationError as e:
        raise ConfigError(f"Resource is invalid: {e!s}")

    _check(properties, ["Instance", "InstanceArn", "InstanceARN"], required=False)

    _check(properties, ["Principal", "Principals"])

    name, principals = _get_value(properties, ["Principal", "Principals"], ensure_list=True)
    for principal_entry in principals:
        _check(principal_entry, ["Type", "PrincipalType"], parent=name)
        _check(principal_entry, ["Id", "PrincipalId", "Ids", "PrincipalIds"], parent=name)

    _check(properties, ["PermissionSet", "PermissionSetArn", "PermissionSets", "PermissionSetArns"])

    name, targets = _get_value(properties, ["Target", "Targets"], ensure_list=True)
    for target_entry in targets:
        _check(target_entry, ["Type", "TargetType"], parent=name)
        _check(target_entry, ["Id", "TargetId", "Ids", "TargetIds"], parent=name)

def validate_config(config, ids: lookup.Ids):
    if not config.instance:
        config.instance = ids.instance_arn
    elif not ids.instance_arn_matches(config.instance):
        LOGGER.warning(f"Config instance {config.instance} does not match {ids.instance_arn}")

    if not (config.groups or config.users):
       raise ConfigError("No principals specified")
    if not config.permission_sets:
        raise ConfigError("No permission sets specified")
    if not (config.ous or config.accounts):
        raise ConfigError("No targets specified")



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import argparse
from collections import namedtuple, OrderedDict
from pathlib import Path
import logging
import sys
import os

import boto3
import aws_error_utils

import click

from aws_sso_lib import lookup
from aws_sso_lib import format as _format
from aws_sso_lib.assignments import Assignment

from .cfn_lib.config import Config, ConfigError, validate_config, GenerationConfig
from .cfn_lib import resources, templates, macro
from .cfn_lib import utils as cfn_utils

from .utils import configure_logging

Input = namedtuple("Input", ["base_path", "stem", "assignments", "permission_sets", "base_template"])

TemplateProcessInput = namedtuple("TemplateProcessInput", [
    "base_path",
    "base_stem",
    "base_template",
    "generation_config",
    "max_stack_resources",
    "items"
])
TemplateProcessInputItem = namedtuple("TemplateProcessInput", [
    "stem",
    "resource_collection",
])

LOGGER = logging.getLogger(__name__)

def param_loader(ctx, param, value):
    if not value:
        return {}
    d = []
    for p in value.split(","):
        if "=" in p:
            d.append((p.split("=", 1)))
        else:
            d.append((p, None))
    return dict(d)

@click.command("cfn")
@click.argument("config_file", required=True, nargs=-1, type=click.File("r"))

@click.option("--macro", is_flag=True, help="Process templates with macro")

@click.option("--profile", metavar="NAME", help="AWS profile to use to retrieve SSO instance and/or accounts from OUs")
@click.option("--sso-instance", "--ins", metavar="ARN", help="If not provided, will be retrieved from your account")

@click.option("--template-file-suffix", metavar="SUFFIX", help="Output template suffix")
@click.option("--output-dir", metavar="DIR", help="Directory for templates")

@click.option("--base-template-file", type=click.File("r"), help="Base template to build from")
@click.option("--template-parameters", callback=param_loader, metavar="PARAMS", help="String-type parameters on the template")

@click.option("--lookup-names/--no-lookup-names", default=False, help="Look up names for principals, permission sets, and accounts")

@click.option("--num-child-stacks", type=int, metavar="NUM", help="Fix the number of child stacks (0 or positive integer)")
@click.option("--max-assignments-allocation", type=int, metavar="NUM", help="Fix a nonzero number of child stacks based on expected max number of assignments")
@click.option("--default-session-duration", metavar="DUR", help="ISO8601 duration for PermissionSets without session duration set")

@click.option("--max-resources-per-template", type=int, metavar="NUM")
@click.option("--max-concurrent-assignments", type=int, metavar="NUM")

@click.option("--assignments-csv", type=click.File("w"), help="Output file name to store CSV of generated assignments")
@click.option("--assignments-csv-only", is_flag=True, help="With --assignments-csv, skip template generation")

@click.option("--verbose", "-v", count=True)
def generate_template(
        config_file,
        macro,
        profile,
        sso_instance,
        template_file_suffix,
        output_dir,
        base_template_file,
        template_parameters,
        lookup_names,
        num_child_stacks,
        max_assignments_allocation,
        default_session_duration,
        max_resources_per_template,
        max_concurrent_assignments,
        assignments_csv,
        assignments_csv_only,
        verbose):
    """Generate CloudFormation templates with AWS SSO assignments."""

    configure_logging(LOGGER, verbose)

    if macro and base_template_file:
        raise click.UsageError("--base-template-file not allowed with --macro")
    if macro and template_parameters:
        raise click.UsageError("--template-parameters not allowed with --macro")

    if assignments_csv_only and not assignments_csv:
        raise click.UsageError("Missing --assignments-csv")

    session = boto3.Session(profile_name=profile)

    ids = lookup.Ids(session, sso_instance, identity_store_id=None)

    cache = {}

    if lookup_names:
        principal_name_fetcher = cfn_utils.get_principal_name_fetcher(session, ids, cache)
        permission_set_name_fetcher = cfn_utils.get_permission_set_name_fetcher(session, ids, cache)
        target_name_fetcher = cfn_utils.get_target_name_fetcher(session, ids, cache)
    else:
        principal_name_fetcher = None
        permission_set_name_fetcher = None
        target_name_fetcher = None

    generation_config = GenerationConfig(
        ids,
        principal_name_fetcher=principal_name_fetcher,
        permission_set_name_fetcher=permission_set_name_fetcher,
        target_name_fetcher=target_name_fetcher
    )

    generation_config.set(
        max_resources_per_template=max_resources_per_template,
        max_concurrent_assignments=max_concurrent_assignments,
        max_assignments_allocation=max_assignments_allocation,
        num_child_stacks=num_child_stacks,
        default_session_duration=default_session_duration,
    )

    if not template_file_suffix:
        template_file_suffix = ".yaml"
    elif not template_file_suffix.endswith(".yaml"):
        template_file_suffix = template_file_suffix + ".yaml"

    if base_template_file:
        base_template = cfn_utils.load_yaml(base_template_file)
        base_template_path = Path(base_template_file.name).resolve()
        prev_len = len(config_file)
        config_file = [c for c in config_file if Path(c.name).resolve() != base_template_path]
        if len(config_file) != prev_len:
            LOGGER.debug("Removed base template file from list of config files")
    else:
        base_template = None

    if macro:
        template_process_inputs = process_macro(
            config_file=config_file,
            session=session,
            ids=ids,
            template_file_suffix=template_file_suffix,
            output_dir=output_dir,
            base_generation_config=generation_config,
        )
    else:
        template_process_inputs = process_config(
            config_file=config_file,
            session=session,
            ids=ids,
            template_file_suffix=template_file_suffix,
            output_dir=output_dir,
            base_template=base_template,
            base_generation_config=generation_config,
        )

    if not assignments_csv_only:
        templates_to_write = process_templates(
            template_process_inputs=template_process_inputs,
            template_file_suffix=template_file_suffix,
        )
        write_templates(templates_to_write)

    if assignments_csv:
        write_csv(template_process_inputs, assignments_csv, generation_config)

def process_config(
    config_file,
    session,
    ids,
    template_file_suffix,
    output_dir,
    base_template,
    base_generation_config):
    template_process_inputs = {}

    for config_file_fp in config_file:
        LOGGER.info(f"Loading config file {config_file_fp.name}")
        config_file_path = Path(config_file_fp.name)
        if output_dir:
            base_path = Path(output_dir)
        else:
            base_path = config_file_path.parent / "templates"
        stem = config_file_path.stem

        data = cfn_utils.load_yaml(config_file_fp)
        LOGGER.debug(f"Config file contents:\n{cfn_utils.dump_yaml(data)}")

        config = Config()
        config.load(data)

        generation_config = base_generation_config.copy()
        generation_config.load(data)

        LOGGER.debug(f"generation_config: {generation_config!s}")

        try:
            validate_config(config, ids)
        except ConfigError as e:
            LOGGER.fatal(f"{e!s} in {config_file_path}")
            sys.exit(1)

        cache = {}
        ou_fetcher = lambda ou, recursive: lookup.lookup_accounts_for_ou(session, ou,
            recursive=recursive,
            cache=cache,
            exclude_org_mgmt_acct=True)

        resource_collection = resources.get_resources_from_config(
            config,
            ou_fetcher=ou_fetcher)

        LOGGER.info(f"Generated {len(resource_collection.assignments)} assignments")

        max_stack_resources = generation_config.get_max_number_of_child_stacks(resource_collection.num_resources)

        template_process_inputs[config_file_path] = TemplateProcessInput(
            base_path=base_path,
            base_stem=stem,
            base_template=base_template,
            generation_config=generation_config,
            max_stack_resources=max_stack_resources,
            items=[TemplateProcessInputItem(
                stem=stem,
                resource_collection=resource_collection
            )]
        )
    return template_process_inputs

def process_macro(
        config_file,
        session,
        ids,
        template_file_suffix,
        output_dir,
        base_generation_config):

    template_process_inputs = {}

    for config_file_fp in config_file:
        LOGGER.info(f"Loading template file {config_file_fp.name}")
        config_file_path = Path(config_file_fp.name)
        if output_dir:
            base_path = Path(output_dir)
        else:
            base_path = config_file_path.parent / "templates"
        stem = config_file_path.stem

        input_template = cfn_utils.load_yaml(config_file_fp)
        LOGGER.debug(f"Input template:\n{cfn_utils.dump_yaml(input_template)}")

        generation_config = base_generation_config.copy()

        LOGGER.info("Extracting resources from template")
        base_template, max_stack_resources, resource_collection_dict = macro.process_template(input_template,
                session=session,
                ids=ids,
                generation_config=generation_config,
                generation_config_template_priority=False)

        num_assignments = sum(len(rc.assignments) for rc in resource_collection_dict.values())
        LOGGER.info(f"Generated {num_assignments} assignments")

        template_process_inputs[config_file_path] = TemplateProcessInput(
            base_path=base_path,
            base_stem=stem,
            base_template=base_template,
            generation_config=generation_config,
            max_stack_resources=max_stack_resources,
            items=[TemplateProcessInputItem(
                stem=resource_name,
                resource_collection=resource_collection
            ) for resource_name, resource_collection in resource_collection_dict.items()]
        )

    return template_process_inputs

def process_templates(
        template_process_inputs,
        template_file_suffix):
    templates_to_write = {}

    for name, template_process_input in template_process_inputs.items():
        LOGGER.info(f"Generating templates for {name}")
        parent_template_to_write = template_process_input.base_template or {}
        all_children = []
        for template_process_input_item in template_process_input.items:
            num_parent_resources = len(parent_template_to_write.get("Resources", {})) + template_process_input.max_stack_resources

            parent_template = templates.resolve_templates(
                template_process_input_item.resource_collection.assignments,
                template_process_input_item.resource_collection.permission_sets,
                generation_config=template_process_input.generation_config,
                num_parent_resources=num_parent_resources,
            )

            template_collection = parent_template.get_templates(
                template_process_input.base_path,
                ".",
                template_process_input_item.stem,
                template_file_suffix,
                base_template=parent_template_to_write,
                parameters=None,
                generation_config=template_process_input.generation_config,
                path_joiner=os.path.join

            )
            parent_template_to_write = template_collection.parent.template
            LOGGER.debug(f"Intermediate parent template\n{cfn_utils.dump_yaml(parent_template_to_write)}")

            all_children.extend(template_collection.children)

        templates_to_write[name] = templates.TemplateCollection(
            parent=templates.WritableTemplate(
                path=os.path.join(
                    template_process_input.base_path,
                    f"{template_process_input.base_stem}{template_file_suffix}"),
                template=parent_template_to_write,
            ),
            children=all_children
        )
    return templates_to_write

def write_templates(templates_to_write):
    for name, template_collection_to_write in templates_to_write.items():
        parent_path = template_collection_to_write.parent.path
        parent_data = template_collection_to_write.parent.template

        for child_path, child_data in template_collection_to_write.children:
            LOGGER.info(f"Writing child template at path {child_path}")
            Path(child_path).parent.mkdir(parents=True, exist_ok=True)
            with open(child_path, "w") as fp:
                cfn_utils.dump_yaml(child_data, fp)

        LOGGER.info(f"Writing template for {name} at path {parent_path}")
        Path(parent_path).parent.mkdir(parents=True, exist_ok=True)
        with open(parent_path, "w") as fp:
            cfn_utils.dump_yaml(parent_data, fp)


def write_csv(template_process_inputs, assignments_csv, generation_config):
    LOGGER.info(f"Writing assignments CSV to {assignments_csv.name}")
    header_fields = Assignment._fields + ("source_ou", )
    assignments_csv.write(",".join(header_fields) + "\n")
    for template_process_input in template_process_inputs.values():
        for template_process_input_item in template_process_input.items:
            for assignment in template_process_input_item.resource_collection.assignments:
                source_ou = assignment.target.source_ou or ""
                assignment_tuple = assignment.get_assignment(
                    principal_name_fetcher=generation_config.principal_name_fetcher,
                    permission_set_name_fetcher=generation_config.permission_set_name_fetcher,
                    target_name_fetcher=generation_config.target_name_fetcher
                )
                assignments_csv.write(",".join(v or "" for v in (assignment_tuple + (source_ou,)) ) + "\n")


if __name__ == "__main__":
    generate_template(prog_name="python -m aws_sso_util.cfn")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import re
from collections import namedtuple
import logging

import boto3
import aws_error_utils

import click

from aws_sso_lib.lookup import Ids
from aws_sso_lib.assignments import _list_assignments, Assignment

from .utils import configure_logging

LOGGER = logging.getLogger(__name__)

def get_principal_filter(group_values, user_values):
    def filter(type, id, name):
        if not (group_values or user_values):
            return True
        if type == "GROUP":
            for value in group_values:
                if id == value or re.search(value, name):
                    return True
        elif type == "USER":
            for value in user_values:
                if id == value or re.search(value, name):
                    return True
        else:
            raise ValueError(f"Unknown principal type {type}")
    return filter

def get_permission_set_filter(values):
    def filter(arn, name):
        if not values:
            return True
        for value in values:
            if arn == value:
                return True
            if arn.split("/", 1)[1] == value:
                return True
            if arn.split("/", 2)[2] == value:
                return True
            if re.search(value, name):
                return True
        return False
    return filter

def get_target_filter(values):
    def filter(type, id, name):
        if type != "AWS_ACCOUNT":
            return True
        if not values:
            return True
        for value in values:
            if id.startswith(value) or id.endswith(value) or re.search(value, name):
                return True
        return False
    return filter

@click.command()
@click.option("--instance-arn", "--ins", metavar="ARN")
@click.option("--identity-store-id", "--id-store", metavar="ID")

@click.option("--profile", metavar="PROFILE_NAME", help="Use a specific AWS profile")

@click.option("--group", "-g", "group_values", multiple=True, default=[], metavar="GROUP", help="Group ID or display name")
@click.option("--user", "-u", "user_values", multiple=True, default=[], metavar="USER", help="User ID or user name")
@click.option("--permission-set", "-p", "permission_set_values", multiple=True, default=[], metavar="PERM_SET", help="Permission set ARN, ID, or name")
@click.option("--account", "-a", "account_values", multiple=True, default=[], metavar="ACCOUNT", help="Account ID or name")
@click.option("--ou", "ou_values", multiple=True, default=[], metavar="ID", help="Organizations OU or root ID")
@click.option("--ou-recursive/--ou-not-recursive", help="Include accounts of child OUs when using --ou")

@click.option("--lookup-names/--no-lookup-names", default=True, help="Look up names for principals, permission sets, and accounts")

@click.option("--show-id/--hide-id", default=False, help="Print SSO instance/identity store id being used")
@click.option("--separator", "--sep", default=",", metavar="SEP", help="Field separator for output")
@click.option("--header/--no-header", default=True, help="Include or supress the header row")
@click.option("--arn-style", type=click.Choice(["arn", "id"]), default="arn", envvar="AWS_SSO_UTIL_ASSIGNMENTS_ARN_STYLE", help="Full ARN or only ID")
@click.option("--verbose", "-v", count=True)
def assignments(
        instance_arn,
        identity_store_id,
        profile,
        group_values,
        user_values,
        permission_set_values,
        account_values,
        ou_values,
        ou_recursive,
        lookup_names,
        show_id,
        separator,
        header,
        arn_style,
        verbose):
    """Retrieve assignments from AWS SSO

    You can filter by principal, permission set, and target."""

    configure_logging(LOGGER, verbose)

    principal = None
    principal_filter = get_principal_filter(group_values, user_values)

    ps_pattern = r"^(arn:aws:sso:::permissionSet/)?((sso)?ins-[0-9a-f]+/)?ps-[0-9a-f]+$"
    if all(re.match(ps_pattern, ps) for ps in permission_set_values):
        LOGGER.debug(f"Using specific permission set ids")
        permission_set = permission_set_values
        permission_set_filter = None
    else:
        if permission_set_values:
            LOGGER.debug("Filtering permission sets")
        permission_set = None
        permission_set_filter = get_permission_set_filter(permission_set_values)

    if account_values and all(re.match(r"^\d{12}$", a) for a in account_values):
        if ou_values:
            LOGGER.debug(f"Using specific accounts and OUs")
        else:
            LOGGER.debug(f"Using specific accounts")
        target = account_values + ou_values
        target_filter = None
    elif account_values and ou_values:
        raise click.UsageError("Cannot provide account names and OUs")
    elif ou_values:
        target = ou_values
        target_filter =  None
    else:
        LOGGER.debug("Filtering account values")
        target = None
        target_filter = get_target_filter(account_values)

    session = boto3.Session(profile_name=profile)

    ids = Ids(lambda: session, instance_arn, identity_store_id)
    ids.print_on_fetch = show_id

    assignments_iterator = _list_assignments(
        session,
        ids,
        principal=principal,
        principal_filter=principal_filter,
        permission_set=permission_set,
        permission_set_filter=permission_set_filter,
        target=target,
        target_filter=target_filter,
        get_principal_names=lookup_names,
        get_permission_set_names=lookup_names,
        get_target_names=lookup_names,
        ou_recursive=ou_recursive)

    if header:
        fields = list(Assignment._fields)
        if arn_style == "id":
            fields[fields.index("instance_arn")] = "instance_id"
            fields[fields.index("permission_set_arn")] = "permission_set_id"
        print(separator.join(fields))

    for assignment in assignments_iterator: #lookup_assignments(session, ids, principal_filter, permission_set_filter, target_filter):
        if arn_style == "id":
            assignment = assignment._replace(instance_arn=assignment.instance_arn.split("/", 1)[-1])
            assignment = assignment._replace(permission_set_arn=assignment.permission_set_arn.split("/", 2)[-1])
        print(separator.join(v or "" for v in assignment))

if __name__ == "__main__":
    assignments(prog_name="python -m aws_sso_util.assignments")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import argparse
import sys
import os
from collections import namedtuple
import logging

import boto3
from botocore.credentials import JSONFileCache

import click

from aws_sso_lib import lookup as _lookup
from aws_sso_lib import format as _format

from .utils import configure_logging, Printer

LOGGER = logging.getLogger(__name__)

IDS_CACHE_DIR = os.path.expanduser(
    os.path.join("~", ".aws", "cli", "cache")
)

@click.command("lookup")
@click.argument("type", type=click.Choice(["instance", "identity-store", "group", "user", "permission-set"]))
@click.argument("value", nargs=-1)

@click.option("--instance-arn", "--ins", metavar="ARN")
@click.option("--identity-store-id", "--id-store", metavar="ID")

@click.option("--profile", metavar="PROFILE_NAME", help="Use a specific AWS profile")

@click.option("--error-if-not-found", "-e", is_flag=True, help="Do not continue if an entity is not found")

@click.option("--show-id/--hide-id", default=False, help="Print SSO instance/identity store id being used")
@click.option("--separator", "--sep", metavar="SEP", help="Field separator for output")
@click.option("--header/--no-header", help="Include or supress the header row")

@click.option("--permission-set-style", type=click.Choice(["arn", "id"]), default="arn", help="Full ARN or only ID")
@click.option("--verbose", "-v", count=True)
def lookup(
        type,
        value,
        instance_arn,
        identity_store_id,
        profile,
        error_if_not_found,
        show_id,
        separator,
        header,
        permission_set_style,
        verbose):
    """Look up names and ids in AWS SSO"""
    configure_logging(LOGGER, verbose)

    session = boto3.Session(profile_name=profile)

    cache = JSONFileCache(IDS_CACHE_DIR)

    ids = _lookup.Ids(session, instance_arn, identity_store_id, cache=cache)
    ids.print_on_fetch = show_id

    HEADER_FIELDS = {
    "group": ["Name", "Id"],
    "user": ["Name", "Id"],
    "permission-set": ["Name", permission_set_style.upper()],
}

    printer = Printer(
        separator=separator,
        default_separator=" ",
        header_fields=HEADER_FIELDS.get(type),
        disable_header=header,
    )

    try:
        if type == "instance":
            ids.print_on_fetch = False
            print(ids.instance_arn)
        elif type == "identity-store":
            ids.print_on_fetch = False
            print(ids.identity_store_id)
        elif type in "group":
            if not value:
                raise click.UsageError("Group name is required")
            lookup_groups(session, ids, value, printer, error_if_not_found=error_if_not_found)
        elif type == "user":
            if not value:
                raise click.UsageError("User name is required")
            lookup_users(session, ids, value, printer, error_if_not_found=error_if_not_found)
        elif type == "permission-set":
            if not value:
                raise click.UsageError("Permission set name is required")
            if len(value) == 1 and value[0] == ":all":
                lookup_all_permission_sets(session, ids, printer, permission_set_style=permission_set_style)
            else:
                lookup_permission_sets(session, ids, value, printer,
                    permission_set_style=permission_set_style,
                    error_if_not_found=error_if_not_found)

    except _lookup.LookupError as e:
        print(e, file=sys.stderr)
        sys.exit(1)

def lookup_groups(session, ids, groups, printer: Printer, *, error_if_not_found):
    printer.print_header_before()
    for value in groups:
        try:
            group_name = value
            group_id = _lookup.lookup_group_by_name(session, ids, group_name)["GroupId"]
        except _lookup.LookupError as e:
            if error_if_not_found:
                printer.print_after()
                print("Group {} not found".format(group_name), file=sys.stderr)
                sys.exit(1)
            group_id = "NOT_FOUND"
        printer.add_row((group_name, group_id))
    printer.print_after()

def lookup_users(session, ids, users, printer: Printer, *, error_if_not_found):
    printer.print_header_before()
    for value in users:
        try:
            user_name = value
            user_id = _lookup.lookup_user_by_name(session, ids, user_name)["UserId"]
        except _lookup.LookupError as e:
            if error_if_not_found:
                printer.print_after()
                print("User {} not found".format(user_name), file=sys.stderr)
                sys.exit(1)
            user_id = "NOT_FOUND"
        printer.add_row((user_name, user_id))
    printer.print_after()

def lookup_permission_sets(session, ids, permission_sets, printer: Printer, *, permission_set_style, error_if_not_found):
    cache = {}
    printer.print_header_before()
    for value in permission_sets:
        try:
            if any(value.startswith(v) for v in ["arn", "ssoins-", "ins-", "ps-"]):
                permission_set = _lookup.lookup_permission_set_by_id(session, ids, value, cache=cache)
            else:
                permission_set = _lookup.lookup_permission_set_by_name(session, ids, value, cache=cache)
            permission_set_name = permission_set["Name"]
            permission_set_arn = permission_set["PermissionSetArn"]
        except _lookup.LookupError as e:
            if error_if_not_found:
                printer.print_after()
                print("Permission set {} not found".format(value), file=sys.stderr)
                sys.exit(1)
            if any(value.startswith(v) for v in ["arn", "ssoins-", "ins-", "ps-"]):
                permission_set_arn = _format.format_permission_set_arn(ids, value)
                permission_set_name = "UNKNOWN"
            else:
                permission_set_arn = "UNKNOWN"
                permission_set_name = value
        if permission_set_style == "id":
            permission_set_arn = permission_set_arn.rsplit("/", 1)[1]
        printer.add_row((permission_set_name, permission_set_arn))
    printer.print_after()

def lookup_all_permission_sets(session, ids, printer: Printer, *, permission_set_style):
    cache = {}
    printer.print_header_before()
    sso = session.client("sso-admin")
    paginator = sso.get_paginator("list_permission_sets")
    for ind, response in enumerate(paginator.paginate(InstanceArn=ids.instance_arn)):
        LOGGER.debug(f"ListPermissionSets page {ind+1}: {', '.join(response['PermissionSets'])}")
        for permission_set_arn in response["PermissionSets"]:
            permission_set = _lookup.lookup_permission_set_by_id(session, ids, permission_set_arn, cache=cache)
            permission_set_name = permission_set["Name"]
            permission_set_arn = permission_set["PermissionSetArn"]
            if permission_set_style == "id":
                permission_set_arn = permission_set_arn.rsplit("/", 1)[1]
            printer.add_row((permission_set_name, permission_set_arn))
    printer.print_after()

if __name__ == "__main__":
    lookup(prog_name="python -m aws_sso_util.lookup")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import logging.handlers

from aws_sso_lib.config import find_instances, SSOInstance

def configure_logging(logger, verbose, **config_args):
    if verbose in [False, None]:
        verbose = 0
    elif verbose == True:
        verbose = 1

    logging.basicConfig(**config_args)

    aws_sso_util_logger = logging.getLogger("aws_sso_util")
    root_logger = logging.getLogger()

    if verbose == 0:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(message)s"))
        logger.addHandler(handler)
        logger.propagate = False
        logger.setLevel(logging.INFO)
    elif verbose == 1:
        logger.setLevel(logging.DEBUG)
        aws_sso_util_logger.setLevel(logging.INFO)
    elif verbose == 2:
        logger.setLevel(logging.DEBUG)
        aws_sso_util_logger.setLevel(logging.DEBUG)
        root_logger.setLevel(logging.INFO)
    elif verbose >= 3:
        logger.setLevel(logging.DEBUG)
        aws_sso_util_logger.setLevel(logging.DEBUG)
        root_logger.setLevel(logging.DEBUG)

class GetInstanceError(Exception):
    pass

def get_instance(sso_start_url, sso_region, sso_start_url_vars=None, sso_region_vars=None):
    instances, specifier, all_instances = find_instances(
        profile_name=None,
        profile_source=None,
        start_url=sso_start_url,
        start_url_source="CLI input",
        region=sso_region,
        region_source="CLI input",
        start_url_vars=sso_start_url_vars,
        region_vars=sso_region_vars
    )

    if not instances:
        if all_instances:
            raise GetInstanceError(
                f"No AWS SSO config matched {specifier.to_str(region=True)} " +
                f"from {SSOInstance.to_strs(all_instances)}")
        else:
            GetInstanceError("No AWS SSO config found")

    if len(instances) > 1:
        GetInstanceError(f"Found {len(instances)} SSO configs, please specify one: {SSOInstance.to_strs(instances)}")

    return instances[0]

class Printer:
    def __init__(self, *,
            separator,
            default_separator,
            header_fields,
            disable_header=False,
            skip_repeated_values=False,
            sort_key=None,
            printer=None):
        self.separator = separator
        self.default_separator = default_separator
        self._sep = separator if separator is not None else default_separator
        self._header_sep = separator if separator is not None else " " * len(default_separator)
        self._justify = separator is None

        self.sort_key = sort_key

        self.header_fields = header_fields
        self.disable_header = disable_header

        self.skip_repeated_values = skip_repeated_values

        self.print_along = self.separator and not self.sort_key
        self.rows = [] if not self.print_along else None

        self.printer = printer or print

    def print_header_before(self):
        if self.print_along and not self.disable_header:
            self.printer(self._header_sep.join(self.header_fields))

    def add_row(self, row):
        if self.print_along:
            self.printer(self._sep.join(row))
        else:
            self.rows.append(row)

    def _process_row_skip(self, row, prev_row):
        if self.skip_repeated_values is True:
            proc = lambda v, pv: "" if v == pv else v
            return [proc(v, pv) for v, pv in zip(row, prev_row)]
        else:
            proc = lambda s, v, pv: "" if s and v == pv else v
            return [proc(s, v, pv) for s, v, pv in zip(self.skip_repeated_values, row, prev_row)]

    def print_after(self):
        if self.print_along:
            return
        if self.sort_key:
            self.rows.sort(key=self.sort_key)

        if self.disable_header:
            col_widths = [0 for _ in self.header_fields]
        else:
            col_widths = [len(h) for h in self.header_fields]

        for row in self.rows:
            col_widths = [max(cw, len(v)) for cw, v in zip(col_widths, row)]

        def just(row):
            if not self._justify:
                return row
            else:
                return [v.ljust(cw) for cw, v in zip(col_widths, row)]

        if not self.disable_header:
            self.printer(self._header_sep.join(just(self.header_fields)))

        first_loop = True
        prev_row = None
        for row in self.rows:
            if not first_loop and self.skip_repeated_values:
                row_to_print = self._process_row_skip(row, prev_row)
            else:
                row_to_print = row

            self.printer(self._sep.join(just(row_to_print)))

            prev_row = row
            first_loop = False



import logging
import re
import sys
from collections import namedtuple

import click

from aws_sso_lib.sso import list_available_roles, login

from .utils import configure_logging, get_instance, GetInstanceError, Printer

LOGGER = logging.getLogger(__name__)

HEADER_FIELDS = {
    "id": "Account ID",
    "name": "Account name",
    "role": "Role name"
}

@click.command()
@click.option("--sso-start-url", "-u", metavar="URL", help="Your AWS SSO start URL")
@click.option("--sso-region", metavar="REGION", help="The AWS region your AWS SSO instance is deployed in")
@click.option("--account", "-a", "account_values", multiple=True, default=[], metavar="ACCOUNT", help="List roles for a specific account ID, can be specified multiple times")
@click.option("--role-name", "-r", "role_name_patterns", multiple=True, default=[], metavar="NAME_REGEX", help="Filter roles by a regular expression, can be specified multiple times")
@click.option("--separator", "--sep", metavar="SEP", help="Field separator for output")
@click.option("--header/--no-header", default=True, help="Include or supress the header row")
@click.option("--sort-by", type=click.Choice(["id,role", "name,role", "role,id", "role,name"]), default=None, help="Specify how the output is sorted")
@click.option("--force-refresh", is_flag=True, help="Re-login")
@click.option("--verbose", "-v", count=True)
def roles(
        sso_start_url,
        sso_region,
        account_values,
        role_name_patterns,
        separator,
        header,
        sort_by,
        force_refresh,
        verbose):
    """List your available accounts and roles.

    --sso-start-url and --sso-region are not needed if a single value can be found for them in your ~/.aws/config
    or in the environment variables AWS_DEFAULT_SSO_START_URL and AWS_DEFAULT_SSO_REGION.

    You can filter the list by providing account IDs and role name patterns.

    """

    configure_logging(LOGGER, verbose)

    if not account_values:
        account_ids = None
        account_filter = lambda id, name: True
    elif all(re.match(r"^\d{12}$", a) for a in account_values):
        account_ids = account_values
        account_filter = lambda id, name: True
    else:
        account_ids = None
        def account_filter(id, name):
            for value in account_values:
                if id.startswith(value) or id.endswith(value) or re.search(value, name):
                    return True
            return False

    if sort_by:
        sort_by_keys = sort_by.split(",")
    elif not separator:
        sort_by_keys = ("name", "role")
    else:
        sort_by_keys = None

    if not sort_by_keys:
        header_field_keys = ("name", "id", "role")
    elif sort_by_keys[0] == "id":
        header_field_keys = ("id", "name", "role")
    elif sort_by_keys[0] == "name":
        header_field_keys = ("name", "id", "role")
    elif sort_by_keys[1] == "id":
        header_field_keys = ("role", "id", "name")
    else:
        header_field_keys = ("role", "name", "id")
    header_fields = [HEADER_FIELDS[k] for k in header_field_keys]
    Row = namedtuple("Row", header_field_keys)

    if sort_by_keys:
        sort_key = lambda v: tuple(getattr(v, key) for key in sort_by_keys)
    else:
        sort_key = None

    try:
        instance = get_instance(
            sso_start_url,
            sso_region,
        )
    except GetInstanceError as e:
        LOGGER.fatal(str(e))
        sys.exit(1)

    login(instance.start_url, instance.region, force_refresh=force_refresh)

    printer = Printer(
        separator=separator,
        default_separator=" ",
        sort_key=sort_key,
        header_fields=header_fields,
        disable_header=not header,
        skip_repeated_values=False,
    )
    printer.print_header_before()

    for account_id, account_name, role_name in list_available_roles(instance.start_url, instance.region, account_id=account_ids):
        if not account_filter(account_id, account_name):
            continue
        if role_name_patterns:
            for pattern in role_name_patterns:
                if re.search(pattern, role_name):
                    break
            else:
                continue
        printer.add_row(Row(id=account_id, name=account_name, role=role_name))

    printer.print_after()


if __name__ == "__main__":
    roles(prog_name="python -m aws_sso_util.roles")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import logging
import copy
import json
import math
import datetime
from collections import namedtuple

import boto3

from aws_sso_lib import lookup

from .config import Config, validate_resource, GenerationConfig
from . import resources, templates, utils, cfn_yaml_tags

"""
https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-macros.html
Input:
{
    "region" : "us-east-1",
    "accountId" : "$ACCOUNT_ID",
    "fragment" : { ... },
    "transformId" : "$TRANSFORM_ID",
    "params" : { ... },
    "requestId" : "$REQUEST_ID",
    "templateParameterValues" : { ... }
}

Output:
{
    "requestId" : "$REQUEST_ID",
    "status" : "$STATUS",
    "fragment" : { ... }
}
"""
TRANSFORM_NAME = "AWS-SSO-Util-2020-11-08"
RESOURCE_TYPE = "SSOUtil::SSO::AssignmentGroup"
PERMISSION_SET_RESOURCE_TYPE = "SSOUtil::SSO::PermissionSet"

LOGGER = logging.getLogger(__name__)

def is_macro_template(template):
    if "Transform" not in template:
        return False
    transform = template["Transform"]
    if transform == TRANSFORM_NAME:
        return True
    if isinstance(transform, list) and TRANSFORM_NAME in transform:
        return True
    return False

def process_template(template,
        session,
        ids: lookup.Ids,
        generation_config: GenerationConfig,
        generation_config_template_priority: bool,
        ou_accounts_cache=None):
    base_template = copy.deepcopy(template)

    generation_config.load(base_template.get("Metadata", {}).get("SSO", {}), overwrite=generation_config_template_priority)
    LOGGER.debug(f"generation_config: {generation_config!s}")

    if "Transform" in base_template:
        if base_template["Transform"] == TRANSFORM_NAME:
            del base_template["Transform"]
        else:
            base_template["Transform"] = [t for t in base_template["Transform"] if t != TRANSFORM_NAME]

    for resource_name, resource in base_template["Resources"].items():
        if resource["Type"] == PERMISSION_SET_RESOURCE_TYPE:
            resource["Type"] = "AWS::SSO::PermissionSet"

    resource_keys = [k for k, v in base_template["Resources"].items() if v["Type"] == RESOURCE_TYPE]
    resource_dict = {k: base_template["Resources"].pop(k) for k in resource_keys}
    LOGGER.debug(f"Found AssignmentGroup resources: {', '.join(resource_keys)}")

    configs = {}

    for resource_name, resource in resource_dict.items():
        validate_resource(resource)
        config = Config()
        config.load_resource_properties(resource["Properties"])
        config.resource_name_prefix = resource_name

        if not config.instance:
            config.instance = ids.instance_arn

        configs[resource_name] = config

    resource_collection_dict = {}
    max_stack_resources = 0
    ou_fetcher = lambda ou, recursive: lookup.lookup_accounts_for_ou(session, ou,
            recursive=recursive,
            cache=ou_accounts_cache,
            exclude_org_mgmt_acct=True)

    for resource_name, config in configs.items():
        resource_collection = resources.get_resources_from_config(
            config,
            assignment_metadata={"AssignmentGroupResourceName": resource_name},
            ou_fetcher=ou_fetcher)

        max_stack_resources += generation_config.get_max_number_of_child_stacks(resource_collection.num_resources)

        resource_collection_dict[resource_name] = resource_collection

    return base_template, max_stack_resources, resource_collection_dict

HANDLER_INITIALIZED = False

SESSION = None
IDS = None

LOOKUP_NAMES = None
CHILD_TEMPLATES_IN_YAML = None

MAX_RESOURCES_PER_TEMPLATE = None
MAX_CONCURRENT_ASSIGNMENTS = None
MAX_ASSIGNMENTS_ALLOCATION = None
NUM_CHILD_STACKS = None
DEFAULT_SESSION_DURATION = None

BUCKET_NAME = None
KEY_PREFIX = None

S3_PUT_OBJECT_ARGS = None

def handler_init():
    global HANDLER_INITIALIZED, \
        SESSION, \
        IDS, \
        LOOKUP_NAMES, \
        CHILD_TEMPLATES_IN_YAML, \
        MAX_RESOURCES_PER_TEMPLATE, \
        MAX_CONCURRENT_ASSIGNMENTS, \
        MAX_ASSIGNMENTS_ALLOCATION, \
        NUM_CHILD_STACKS, \
        DEFAULT_SESSION_DURATION, \
        BUCKET_NAME, \
        KEY_PREFIX, \
        S3_PUT_OBJECT_ARGS
    if HANDLER_INITIALIZED:
        return

    logging.getLogger("aws_sso_util").setLevel(getattr(logging, os.environ.get("LOG_LEVEL", "INFO")))
    logging.basicConfig()

    LOGGER.info("Initializing handler")

    SESSION = boto3.Session()
    IDS = lookup.Ids(lambda: SESSION)

    LOOKUP_NAMES = os.environ.get("LOOKUP_NAMES", "false").lower() in ["true", "1"]
    CHILD_TEMPLATES_IN_YAML = os.environ.get("CHILD_TEMPLATES_IN_YAML", "false").lower() in ["true", "1"]

    MAX_RESOURCES_PER_TEMPLATE = int(os.environ["MAX_RESOURCES_PER_TEMPLATE"]) if os.environ.get("MAX_RESOURCES_PER_TEMPLATE") else None
    MAX_CONCURRENT_ASSIGNMENTS = int(os.environ["MAX_CONCURRENT_ASSIGNMENTS"]) if os.environ.get("MAX_CONCURRENT_ASSIGNMENTS") else None
    MAX_ASSIGNMENTS_ALLOCATION = int(os.environ["MAX_ASSIGNMENTS_ALLOCATION"]) if os.environ.get("MAX_ASSIGNMENTS_ALLOCATION") else None
    NUM_CHILD_STACKS = int(os.environ["NUM_CHILD_STACKS"]) if os.environ.get("NUM_CHILD_STACKS") else None
    DEFAULT_SESSION_DURATION = os.environ["DEFAULT_SESSION_DURATION"] if os.environ.get("DEFAULT_SESSION_DURATION") else None

    BUCKET_NAME = os.environ["BUCKET_NAME"]
    KEY_PREFIX = os.environ.get("KEY_PREFIX", "")

    try:
        S3_PUT_OBJECT_ARGS = json.loads(os.environ["S3_PUT_OBJECT_ARGS"]) if os.environ.get("S3_PUT_OBJECT_ARGS") else {}
    except:
        LOGGER.exception("Error parsing S3_PUT_OBJECT_ARGS")

    HANDLER_INITIALIZED = True

def handler(event, context, put_object=None):
    request_id = event["requestId"]
    LOGGER.info(f"Request ID: {request_id}")

    try:
        handler_init()

        input_template = event["fragment"]

        LOGGER.debug(f"Input template:\n{utils.dump_yaml(input_template)}")

        if "Resources" not in input_template:
            raise TypeError(f"{TRANSFORM_NAME} can only be used as a template-level transform")

        lookup_cache = {}

        if LOOKUP_NAMES:
            principal_name_fetcher = utils.get_principal_name_fetcher(SESSION, IDS, lookup_cache)
            permission_set_name_fetcher = utils.get_permission_set_name_fetcher(SESSION, IDS, lookup_cache)
            target_name_fetcher = utils.get_target_name_fetcher(SESSION, IDS, lookup_cache)
        else:
            principal_name_fetcher = None
            permission_set_name_fetcher = None
            target_name_fetcher = None

        generation_config = GenerationConfig(
            IDS,
            principal_name_fetcher=principal_name_fetcher,
            permission_set_name_fetcher=permission_set_name_fetcher,
            target_name_fetcher=target_name_fetcher
        )
        generation_config.set(
            max_resources_per_template=MAX_RESOURCES_PER_TEMPLATE,
            max_concurrent_assignments=MAX_CONCURRENT_ASSIGNMENTS,
            max_assignments_allocation=MAX_ASSIGNMENTS_ALLOCATION,
            num_child_stacks=NUM_CHILD_STACKS,
            default_session_duration=DEFAULT_SESSION_DURATION,
        )

        ou_accounts_cache = {}

        LOGGER.info("Extracting resources from template")
        output_template, max_stack_resources, resource_collection_dict = process_template(input_template,
                session=SESSION,
                ids=IDS,
                generation_config=generation_config,
                generation_config_template_priority=True,
                ou_accounts_cache=ou_accounts_cache)

        num_assignments = sum(len(rc.assignments) for rc in resource_collection_dict.values())
        LOGGER.info(f"Generated {num_assignments} assignments from {len(resource_collection_dict)} resources")

        all_child_templates_to_write = []

        s3_base_path_parts = [
            "templates",
            f"{datetime.datetime.utcnow().isoformat()[:16]}_{request_id}",
        ]
        if KEY_PREFIX:
            s3_base_path_parts.insert(1, KEY_PREFIX)
        s3_base_path = "/".join(s3_base_path_parts)

        for resource_name, resource_collection in resource_collection_dict.items():
            num_parent_resources = len(output_template["Resources"]) + max_stack_resources

            parent_template = templates.resolve_templates(
                resource_collection.assignments,
                resource_collection.permission_sets,
                generation_config=generation_config,
                num_parent_resources=num_parent_resources,
            )

            suffix = ".yaml" if CHILD_TEMPLATES_IN_YAML else ".json"

            template_collection = parent_template.get_templates(
                s3_base_path,
                f"https://s3.amazonaws.com/{BUCKET_NAME}/{s3_base_path}",
                resource_name,
                suffix,
                generation_config=generation_config,
                base_template=output_template,
                path_joiner=lambda *args: "/".join(args)

            )
            output_template = template_collection.parent.template
            LOGGER.debug(f"Intermediate output template:\n{utils.dump_yaml(output_template)}")

            all_child_templates_to_write.extend(template_collection.children)

        LOGGER.info(f"Writing {len(all_child_templates_to_write)} child templates")

        bucket = boto3.resource("s3").Bucket(BUCKET_NAME)

        for child_template_key, child_template in all_child_templates_to_write:
            LOGGER.debug(f"Writing child template {child_template_key}")
            if CHILD_TEMPLATES_IN_YAML:
                content = utils.dump_yaml(child_template)
            else:
                child_template = cfn_yaml_tags.to_json(child_template)
                content = json.dumps(child_template, indent=2)

            put_object_args = copy.deepcopy(S3_PUT_OBJECT_ARGS)

            put_object_args.update({
                "Key": child_template_key,
                "Body": content,
            })
            if "ContentType" not in put_object_args:
                content_type = "text/plain" if CHILD_TEMPLATES_IN_YAML else "application/json"
                put_object_args["ContentType"] = content_type

            if put_object:
                put_object(**put_object_args)
            else:
                bucket.put_object(**put_object_args)

        output_template = cfn_yaml_tags.to_json(output_template)

        LOGGER.debug(f"Final output template:\n{utils.dump_yaml(output_template)}")

        output = {
            "requestId" : request_id,
            "status" : "success",
            "fragment" : output_template,
        }

        return output
    except Exception as e:
        LOGGER.exception(f"An error occurred: {e}")
        output = {
            "requestId" : request_id,
            "status" : "failure",
            "fragment" : {},
            "errorMessage": str(e),
        }




# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aws_sso_util', 'aws_sso_util.cfn_lib']

package_data = \
{'': ['*']}

install_requires = \
['aws-error-utils>=1.0.4,<2.0.0',
 'aws-sso-lib>=1.5.0,<2.0.0',
 'boto3>=1.16.13,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pyyaml>=5.3.1,<6.0.0']

entry_points = \
{'console_scripts': ['aws-sso-util = aws_sso_util.cli:cli']}

setup_kwargs = {
    'name': 'aws-sso-util',
    'version': '3.21.0',
    'description': 'Utilities to make AWS SSO easier',
    'long_description': '# aws-sso-util\n## Making life with AWS SSO a little easier\n\n[AWS SSO](https://aws.amazon.com/single-sign-on/) has some rough edges, and `aws-sso-util` is here to smooth them out, hopefully temporarily until AWS makes it better.\n\n`aws-sso-util` contains utilities for the following:\n* Configuring `.aws/config`\n* Logging in/out\n* AWS SDK support\n* Looking up identifiers\n* CloudFormation\n\nThe underlying Python library for AWS SSO authentication is [`aws-sso-lib`](https://pypi.org/project/aws-sso-lib/), which has useful functions like interactive login, creating a boto3 session for specific a account and role, and the programmatic versions of the `lookup` functions in `aws-sso-util`.\n\n`aws-sso-util` supersedes `aws-sso-credential-process`, which is still available in its original form [here](https://github.com/benkehoe/aws-sso-credential-process).\nRead the updated docs for `aws-sso-util credential-process` [here](docs/credential-process.md).\n\n## Quickstart\n\n1. I recommend you install [`pipx`](https://pipxproject.github.io/pipx/), which installs the tool in an isolated virtualenv while linking the script you need.\n\nMac [and Linux](https://docs.brew.sh/Homebrew-on-Linux):\n```bash\nbrew install pipx\npipx ensurepath\n```\n\nOther:\n```bash\npython3 -m pip install --user pipx\npython3 -m pipx ensurepath\n```\n\n2. Install\n```bash\npipx install aws-sso-util\n```\n\n3. Learn\n```bash\naws-sso-util --help\n```\n\n## Documentation\n\nSee the full docs at [https://github.com/benkehoe/aws-sso-util](https://github.com/benkehoe/aws-sso-util)\n',
    'author': 'Ben Kehoe',
    'author_email': 'ben@kehoe.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/benkehoe/aws-sso-util',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
